#!/usr/bin/python
# -*- coding: utf-8 -*-

class iodClientFinder():
    """
    Methods to find IP-over-DNS traffic in a 
    set of Impacket data packets

    Tar inn samples som bestaar av x pakker. Slik brukt vil ett
    sample kun innholde pakker fra en klient, men samme klienter
    lagres i samme datastruktur.

    Metodene er utvidet til aa akseptere klientidentifikasjon som
    argument.  
"""
    import zlib
    from datetime import datetime
    def __init__(self, config):
        #self.reset()
        # iodFinder is used to analyze samples of n seconds worth
        # of packets. Adjust qps/bps values by this to get proper
        # bytes per second. 

        self.clistore = {}
        self.config = config
        pass

    def reset(self, client):
        # ts of first data point in sample.
        #raise Exception
        #self.sample_start_ts = None
        if self.clistore.has_key(client):
            del self.clistore[client]

    def add(self, packetlist):
        if len(packetlist) == 0:
            raise Exception, "tom pakkeliste"
  
        for hdr, p in packetlist:
            ip = p.child()
            udp = ip.child()
            data = udp.get_data_as_string()
            ipaddr = ip.get_ip_dst()

            #print "adding ipaddr %s" % ipaddr

            labels = self._querylabels(data[12:])
            domainkey = ".".join(labels[-3:])

            tstuple = hdr.getts()
            ts = self.datetime.fromtimestamp(tstuple[0])
            ts = ts.replace(microsecond=tstuple[1])

            if not self.clistore.get(ipaddr):
                self.clistore[ipaddr] = { 'queries': 0,
                           'bytes': 0,
                           'bytes_kolmo': 0,
                           'bytes_kolmo_max': 0,
                           'kolmo_data': [],
                           'mtbq': 0.0, 
                           'mtbq_store': [],
                           'domain_store': {},
                           'first_ts': ts,
                           }


            self.clistore[ipaddr]["bytes"] += len(data)
            self.clistore[ipaddr]["queries"] += 1

            # virker unodvendig aa maatte implementere en full
            # dnspakkeleser. sporrefeiltet er ca. ferdig rundt 30
            # byte ut i pakken, og resten er dominert av
            # svardata.
            if len(data) < 35:
                kolmobasis = ''
            else:
                kolmobasis = data[30:]

            self.clistore[ipaddr]["bytes_kolmo_max"] += len( kolmobasis )
            self.clistore[ipaddr]["kolmo_data"] += [ kolmobasis ]
            
            self.clistore[ipaddr]["mtbq_store"].append( ts )

            try:
                self.clistore[ipaddr]["domain_store"][domainkey] += len(data)
            except KeyError:
                self.clistore[ipaddr]["domain_store"][domainkey] = len(data)

        # sette slutt-tidspunkt for klient i dette samplet.
        self.clistore[ipaddr]["last_ts"] = ts

        # beregne kolmogorov-verdi av konkatineringen av samtlige
        # question-strenger.
        kbasis = "".join(self.clistore[ipaddr]["kolmo_data"])
        kolmobasis_zipped = self.zlib.compress( kbasis )

        if len(kolmobasis) == 0 or self.clistore[ipaddr]["bytes_kolmo_max"] == 0:
            self.clistore[ipaddr]["kolmoratio"] = 0.00
        else:
            self.clistore[ipaddr]["kolmoratio"] = \
                float( len( kolmobasis_zipped ) ) / \
                self.clistore[ipaddr]["bytes_kolmo_max"] 

        del self.clistore[ipaddr]["kolmo_data"]

    def preprocess(self, client):
        # vil bare vaere en, men beholder pga. likhet med
        # iodFinder.
        key = client
        val = self.clistore[client]
        # siste sample kan være ganske puny, få pakker og lite data. ignorer
        # dette i stillhet.
        if val["queries"] < 2:
            return 

        for i in range(0, 1):
            val["mtbq_store"].sort()
            val["mtbq_deltas"] = []

            # n^2 is my friend
            for i in range(1, len( val["mtbq_store"])):
                delta = val["mtbq_store"][i] - val["mtbq_store"][i-1]
#                print dir( delta)
                delta_ms = delta.seconds * 1000 + (delta.microseconds / 1000.) 
                delta = delta_ms / float(1000)
                #print delta.seconds
                #print delta.microseconds
                val["mtbq_deltas"].append( delta )
#            print "sum: ", sum( val["mtbq_deltas"])
#            print "length: ", len( val["mtbq_deltas"])
            delta_sum = sum( val["mtbq_deltas"] )
            numdeltas = len( val["mtbq_deltas"] )
            if numdeltas == 0:
               val["mtbq"] == None
            else:
                val["mtbq"] = float(delta_sum) / numdeltas

            # mean packet size
            val["mps"] = (float(val["bytes"]) / val["queries"] )

            del val["mtbq_store"]
            del val["mtbq_deltas"]

            #
            val["domain_max_bytes"] = 0
            val["domain_max_name"] = None

            for domain, domainbytes in \
              val["domain_store"].items():
                if domainbytes > val["domain_max_bytes"]:
                    val["domain_max_bytes"] = domainbytes
                    val["domain_max_name"] = domain

            val["domain_max_percent"] = \
                (100 * float(val["domain_max_bytes"]) / val["bytes"])

            sample_length = val["last_ts"] - val["first_ts"]
            #sample_length_ms = sample_length.seconds * 1000 + \
            #     sample_length.microseconds
            # milli, mikro
            sample_length_ms = (sample_length.seconds * 1000) + (sample_length.microseconds / float(1000))

            # foreloepig greit aa holde seg paa sekunder, saa
            # faar vi heller leve med avrundingsfeilene.
            val["sample_length"] = float( sample_length_ms ) / 1000.

            val["qps"] = float(val["queries"]) \
                        / val["sample_length"]
            val["bps"] = float(val["bytes"]) \
                        / val["sample_length"]

            del val["domain_max_bytes"]
            del val["domain_store"]

    def report(self, client, header=True):
        keynames = self.clistore.values()[0].keys()
        keynames.pop( keynames.index("last_ts") ) 
        keynames.pop( keynames.index("first_ts")) 

        if header:
            # header
            print "ts\t",
            print "ip\t",
            print "\t".join(keynames)
            #
        # for hver ip observert
        for ip in [ client ]: 
            print "%s\t" % \
                self.clistore[ip]["first_ts"].strftime("%X"), 

            strings = ["ip",
                "domain_max_name", 
                "last_ts","first_ts"]
            print "%s" % ip,
            for key in keynames:
#                print key
                print "\t",
                if key in strings:
                    print self.clistore[ip][key],
                else:
                    print "%.2f" % \
                        float(self.clistore[ip][key]),
            print

    def stats_formatted(self):
        r = self.stats()
        print "%s\t" % self.sample_start_ts,
        print "average\t",
        if False:
            print "%.2f +-%.2f\t" % (r["mtbq"][0], r["mtbq"][1]),
            print "\t\t\t",
            print "%.2f +-%.2f\t" % (r["bytes"][0], r["bytes"][1]),
            print "%.2f +-%.2f\t" % (r["mps"][0], r["mps"][1]),
            print "%.2f +-%.2f\t" % (r["queries"][0], r["queries"][1]),
        else:

            print "%.2f\t" % (r["mtbq"][0]),
            print "\t\t\t",
            print "%.2f\t" % (r["bytes"][0]),
            print "%.2f\t" % (r["mps"][0]),
            print "%.2f\t" % (r["queries"][0]),

        #print "%.2f (stdev: %.2f)\t" % (r["queries"][0], r["queries"][1]),
        print 


    def stats2_header(self):
        r = ["ts"]

        for i in ["mtbq","mps","bps"]:
            for j in ["iodine","mean", "stdev"]:
                r.append( "%s_%s" % (i, j))

        r += ["numclients"]
        r += ["qps_per_cli"]
        return r
        #print "\t".join(r)

    def stats2(self):
        "tid unike_klienter iodine_bytes iodine_x i_y, avg_foo avg_bar, avg_qux"

        r = self.stats()
        iod = self.clistore.get("129.241.208.220") or {}

        res = [ "%s" % self.sample_start_ts ]

        for i in ["mtbq","mps","bps"]:
            iodine_value = iod.get(i)
            if not iodine_value:
                iodine_value = "NA"
            else:
                iodine_value = "%.2f" % float(iodine_value)
            res.append( "%s" % iodine_value)
            res.append( "%.2f" % r.get(i)[0] )
            res.append( "%.2f" % r.get(i)[1] )

        # numclients og qps_per_cli 
        res += [ str(len(self.clistore)) ]
        qps_per_cli = r.get("qps")[0] / float(len(self.clistore))
        res += [ "%.6f" % (qps_per_cli) ]
        
        return res 
        #print "\t".join(res)

    def stats(self):
        """Find sample statistics
        
        mean 
        stdev

        of: 
            mean packet size. 
            mean time between queries. 
            numqueries
            numbytes

        """
        from statlib import stats
        keys = ["mps","mtbq","qps","bps"]
        res = {}
        for name in keys:
            # dra ut data. litt hummer&kanari om det er string
            # eller ei, saa cast alt til float.
            vector = [ float(i[name]) for i in self.clistore.values() ]
            
            if len(vector) == 0:
                raise Exception, "No values found for key %s" % name

            elif len(vector) == 1:
                res[name] = ( vector[0], 0.0 )
            else:
                # median? quantiles?
                res[name] = ( stats.mean( vector ), stats.stdev( vector) )
        return res


    def _querylabels(self, data):
        """Return a list of the labels in the query portion of a
        DNS packet"""
        i = 0
        labels = []
        while True:
            # print querydata[i]
            llen = int("%02x" % ord(data[i]), 16)
            #print type(llen), llen

            if llen == 0:
                break

            s1 = i+1
            s2 = s1 + llen 
            label = data[s1:s2]

            labels.append(label)
            i = i + llen +1
            
        return labels


class iodClientFinderDST(iodClientFinder):
    def __init__(self, config):
        iodClientFinder.__init__(self, config)
        self.persistant_store = {}  
        self.tick = {}

        self.banditstore = {}
        self.sample_stop_ts = None

    def analyze(self, client):
        """
        Gir denne klienten noen advarsel/alarm i dette samplet?
        returnerer True, false.

        Analyze the current sample """
        verbose = True
        veryverbose = True

        data = self.clistore[client]

        # tester som helt klart sier at det IKKE skal gis alarm,
        # men som ikke sier noe om det skal gis alarm.
        if data["bps"] < self.config["bps_minimum"]: #minbytes:
            if veryverbose:
                #self.pprint(data)
                print "non-iod, low data: %s below configured %s" % \
                    (data["bps"], self.config["bps_minimum"])
            return False

        if data["qps"] < self.config["qps_minimum"]:
            if veryverbose:
                print "non-iod, qps of %s below required/configured %s" % \
                    (data["qps"], self.config["qps_minimum"])
            return False

        if data["mps"] < self.config["mps_minimum"]:
            if veryverbose:
                print "non-iod, mean packet size of " + \
                   "%s below required/configured %s" % \
                   (data["mps"], self.config["mps_minimum"])
            return False


        final_result = False
        # tester som sier om det kan vaere iod.
        if data["domain_max_percent"] > self.config["domain_max_percent"]:
            if verbose:
                print "domain_max_percent alarm: " + \
                    "%s more than configured %s" % \
                    (data["domain_max_percent"], self.config["domain_max_percent"])
            final_result = final_result or True

        if final_result is True:
            print self.pprint(data)
            return final_result
        return final_result

    def timely_analysis(self, client, depth, verbose=False):
        """
        Analyse per klient, over tid. Om en klient sitt datasett
        gir alarm, vil det lagret her.  Maalet er aa se etter
        etterfoelgende alarmer, saa det vil fungere aa gruppere per
        bruker 
        """
        # shift
        if not self.tick.has_key(client):
            self.tick[client] = 0
        self.tick[client] = self.tick[client] + 1
        #for i in range(7, 2):
        #if depth > 8:
        #    raise Exception, "not supported"

        if not self.persistant_store.has_key(client):
            # n last lists of suspects.
            self.persistant_store[client] = \
                [None]* depth

        ps = self.persistant_store
        c = client

        if depth == 1:
            return self.analyze(c)

        if verbose:
            print "pretick %-20s %s: %s" % \
               (client, self.tick[client], ps[client])
        # shift right
        # hva om den er to?
        if depth >= 2:
            for i in range(depth-1, 0, -1):
                #print "shift %s" % i
                self.persistant_store[c][i] = self.persistant_store[c][i-1]

        self.persistant_store[c][0] = self.analyze(c)

        if verbose:
            print "   tick %-20s %s: %s" % \
               (client, self.tick[client], ps[client])

        # AND sammen listen over true/false paa de siste n
        # samples.
        warnings = ps[c][0]
        for i in range(0, depth):
            warnings = warnings and ps[c][i]
        #print "final: ", warnings
        return warnings # alarm

    def pprint(self, d):
        import pprint
        pprint.pprint(d)

    def report(self, client, header=True):
        raise Exception, "Skal subklasses"
        alarm = self.timely_analysis(client, \
            self.config["movingaveragedepth"], verbose=True)
        if not alarm:
            return None

        print "alarm: %s " % (client)

        strings = ["ip","domain_max_name"]
        if not self.banditstore.has_key(client):
            self.banditstore[client] = []
        self.banditstore[client] += [ self.clistore[client] ]

        r = "Probable IP-over-DNS client:" + \
            "%s (%.0f%% of %.0f bytes from domain %s)" % \
                (
                client,
                self.clistore[client]["domain_max_percent"],
                self.clistore[client]["bytes"],
                self.clistore[client]["domain_max_name"],
                ),
        print " ".join(r)


