#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
# Generell iodFinder-klasse, subklasses av 
# de andre mer spesifikke.

class iodFinder():
    """
    Methods to find IP-over-DNS traffic in a set of Impacket data
    packets
    """
    import zlib
    from datetime import datetime
    def __init__(self, slen):
        self.reset()
        # iodFinder is used to analyze samples of n seconds worth
        # of packets. Adjust qps/bps values by this to get proper
        # bytes per second. 
        self.sample_length = float(slen)
        pass

    def reset(self):
        # ts of first data point in sample.
        self.sample_start_ts = None
        self.clistore = {}

    def add(self, packetlist):
        for hdr, p in packetlist:
            ip = p.child()
            udp = ip.child()
            data = udp.get_data_as_string()
            ipaddr = ip.get_ip_dst()

            #print "adding ipaddr %s" % ipaddr

            labels = self._querylabels(data[12:])
            domainkey = ".".join(labels[-3:])

            if not self.clistore.get(ipaddr):
                self.clistore[ipaddr] = { 'queries': 0,
                           'bytes': 0,
                           'bytes_kolmo': 0,
                           # mean time between queries
                           'mtbq': 0.0, 
                           'mtbq_store': [],
                           'domain_store': {},
                           }

            #print len(data),
            data_compressed = self.zlib.compress(data)

            self.clistore[ipaddr]["bytes_kolmo"] += len(data_compressed)
            self.clistore[ipaddr]["bytes"] += len(data)
            self.clistore[ipaddr]["queries"] += 1
            
            ts = hdr.getts()[0]

            if not self.sample_start_ts:
                #print "setting first sample ts to ", ts
                self.sample_start_ts = \
                    self.datetime.fromtimestamp(ts)


            self.clistore[ipaddr]["mtbq_store"].append( ts )

            try:
                self.clistore[ipaddr]["domain_store"][domainkey] += len(data)
            except KeyError:
                self.clistore[ipaddr]["domain_store"][domainkey] = len(data)


    def preprocess(self):
        for key, val in self.clistore.items():
            val["mtbq_store"].sort()
            val["mtbq_deltas"] = []

            # n^2 er min venn.
            for i in range(1, len( val["mtbq_store"])):
                delta = val["mtbq_store"][i] - val["mtbq_store"][i-1]
                val["mtbq_deltas"].append( delta )

            delta_sum = sum( val["mtbq_deltas"] )
            numdeltas = len( val["mtbq_deltas"] )
            if numdeltas == 0:
               val["mtbq"] == None
            else:
                val["mtbq"] = float(delta_sum) / numdeltas

            # mean packet size
            val["mps"] = (float(val["bytes"]) / val["queries"] )

            del val["mtbq_store"]
            del val["mtbq_deltas"]

            val["domain_max_bytes"] = 0
            val["domain_max_name"] = None

            for domain, domainbytes in val["domain_store"].items():
                if domainbytes > val["domain_max_bytes"]:
                    val["domain_max_bytes"] = domainbytes
                    val["domain_max_name"] = domain

            val["domain_max_percent"] = \
                (100 * float(val["domain_max_bytes"]) / \
                    val["bytes"])

            val["qps"]=float(val["queries"]) /self.sample_length
            val["bps"]=float(val["bytes"])   /self.sample_length

            del val["domain_max_bytes"]
            del val["domain_store"]

    def report(self, header=True):
        raise Exception, "SUBCLASS ME, PLEASE!"

    def stats_formatted(self):
        r = self.stats()
        print "%s\t" % self.sample_start_ts,
        print "average\t",
        if False:
            print "%.2f +-%.2f\t" % (r["mtbq"][0], r["mtbq"][1]),
            print "\t\t\t",
            print "%.2f +-%.2f\t" % (r["bytes"][0], r["bytes"][1]),
            print "%.2f +-%.2f\t" % (r["mps"][0], r["mps"][1]),
            print "%.2f +-%.2f\t" % (r["queries"][0], r["queries"][1]),
        else:

            print "%.2f\t" % (r["mtbq"][0]),
            print "\t\t\t",
            print "%.2f\t" % (r["bytes"][0]),
            print "%.2f\t" % (r["mps"][0]),
            print "%.2f\t" % (r["queries"][0]),

        print 


    def stats2_header(self):
        """
        Gir en utskrivbar header for stats2() sitt
        innhold.

        Lite brukt.
        """
        r = ["ts"]

        for i in ["mtbq","mps","bps"]:
            for j in ["iodine","mean", "stdev"]:
                r.append( "%s_%s" % (i, j))

        r += ["numclients"]
        r += ["qps_per_cli"]
        return r

    def stats2(self):
        """
        Henter ut en tekstvariant av statistikken
        til et sample.

        Lite brukt.
        
        keys:
        tid unike_klienter iodine_bytes iodine_x i_y, avg_foo
        avg_bar, avg_qux
        """

        r = self.stats()
        iod = self.clistore.get("129.241.208.220") or {}

        res = [ "%s" % self.sample_start_ts ]

        for i in ["mtbq","mps","bps"]:
            iodine_value = iod.get(i)
            if not iodine_value:
                iodine_value = "NA"
            else:
                iodine_value = "%.2f" % float(iodine_value)
            res.append( "%s" % iodine_value)
            res.append( "%.2f" % r.get(i)[0] )
            res.append( "%.2f" % r.get(i)[1] )

        # numclients og qps_per_cli 
        res += [ str(len(self.clistore)) ]
        qps_per_cli=r.get("qps")[0] / float(len(self.clistore))
        res += [ "%.6f" % (qps_per_cli) ]
        
        return res 

    def stats(self):
        """
        Find sample statistics
        
        mean 
        stdev

        of: 
            mean packet size. 
            mean time between queries. 
            numqueries
            numbytes

        """
        from statlib import stats
        keys = ["mps","mtbq","qps","bps"]
        res = {}
        for name in keys:
            # dra ut data. litt hummer&kanari om det er string
            # eller ei, så cast alt til float.
            vector = [ float(i[name]) for i in self.clistore.values() ]
            
            if len(vector) == 0:
                raise Exception, "No values found for key %s" % \
                    name

            elif len(vector) == 1:
                res[name] = ( vector[0], 0.0 )
            else:
                res[name] = ( 
                    stats.mean( vector ), 
                    stats.stdev( vector)
                )
        return res


    def _querylabels(self, data):
        """
        Trekk ut en liste av labels i Question-delen av et
        DNS-svar.
        """
        i = 0
        labels = []
        while True:
            # print querydata[i]
            llen = int("%02x" % ord(data[i]), 16)
            #print type(llen), llen

            if llen == 0:
                break

            s1 = i+1
            s2 = s1 + llen 
            label = data[s1:s2]

            labels.append(label)
            i = i + llen +1
            
        return labels

