#!/usr/bin/python
# -*- coding: utf-8 -*-
# 
# Kode for å analysere uniformt samplede datasett
# for klienter som sender statistisk svært mye data.
#

from iodFinder import iodFinder

class iodFinderBW(iodFinder):
    def __init__(self, slen):
        iodFinder.__init__(self, slen)
        # three last lists of suspects.
        self.persistant_store = [ [], [], [] ] 
        self.tick = 0

    def analyze(self):
        """
        Analyze the current sample for high bandwidth
        clients
        """
        r = self.stats()
        smean = r["bps"][0]
        sstdev = r["bps"][1] 
        outlier_limit = smean + 3 * sstdev

        print "Sample average:\t %.4f bps" % smean
        print "Sample stdev:\t%.4f " % sstdev
        print "Outlier limit:\t%.4f bps" % outlier_limit

        suspects = []
        # for hver ip observert
        for ip in self.clistore.keys():
            if self.clistore[ip]["bps"] < outlier_limit:
                continue
            suspects.append(ip)
        return {'suspects': set(suspects), 
                'mean': smean, 'stdev': sstdev}

    def timely_analysis(self):
        self.tick = self.tick + 1
        self.persistant_store[2] = \
            self.persistant_store[1]
        self.persistant_store[1] = \
            self.persistant_store[0]
        self.persistant_store[0] = \
            self.analyze()["suspects"]

        ps = self.persistant_store
        bandits = \
        ps[0].intersection(ps[1]).intersection(ps[2])
        return bandits

    def report(self, header=True):
        # første variant uten snitting over tid:
        bandits = self.analyze()["suspects"]

        # andre variant:
        #bandits = self.timely_analysis()

        if len(bandits) == 0:
            #print "nothing found"
            return
        else:
            for ip in bandits:
                strings = ["ip","domain_max_name"]
                print "Probable IP-over-DNS client: "
                print "%s (%.f bytes received in sample)" % \
                    ( ip,
                    self.clistore[ip]["bytes"],
                    ),
                print
