#!/usr/bin/python
# .: coding: utf-8 :.
#
# apt-get install python-pcapy python-impacket


import os, sys
import pcapy
from impacket.ImpactDecoder import EthDecoder
from impacket.ImpactDecoder import LinuxSLLDecoder, UDPDecoder

from ClientStorage import ClientStorage
from iodClientFinder import iodClientFinderDST, iodClientFinder

class Decoder():
    def __init__(self, pcapObj):
        datalink = pcapObj.datalink()
        if pcapy.DLT_EN10MB == datalink:
            self.decoder = EthDecoder()
        elif pcapy.DLT_LINUX_SLL == datalink:
            self.decoder = LinuxSLLDecoder()
        else:
            raise Exception("Datalink type not supported: " \
                % datalink)

        self.pcap = pcapObj
        self.connections = {}
        self.UDPDecoder = UDPDecoder()

    def start(self):
        self.pcXap.loop(0, self.packetHandler)

def packetHandler(hdr, data):
    global clistore, eagle
    p = EthDecoder()
    p = p.decode(data)

    # hopp over ipv6-pakkene, impacket st0tter det ikke.
    ptype = "%x" % ord(p.get_data_as_string()[0])
    if ptype == "60":
        return

    ip = p.child()
    udp = ip.child()
    data = udp.get_data_as_string()

    dns1 = ord(data[2]) 
    dns2 = ord(data[3])

    dnscode = "%x%x" % (dns1, dns2)

    # standard query response, no error
    if not dnscode == "8180": 
        #print "not a dns response"
        return

    client = ip.get_ip_dst()
    res = clistore.add( client, hdr, p )

    if res:
        #print "Yielded %s packets for client %s" % (res, client)
        eagle.add( clistore.get_data(client) )
        eagle.preprocess(client)
        eagle.report(client) # , header=False)

        eagle.reset(client)
        clistore.expunge(client)


if __name__ == "__main__":
    config = { 
        'domain_max_percent': 98, 
        'dev': 'eth1',
        'numpackets': 70,
    }

    config_rates_ssh = {
        # minimum antall bytes per sekund i snitt over samplet.
        'bps_minimum': 560, 
        'qps_minimum': 2.27, # empirisk
        'mps_minimum': 140, # empirisk
        'movingaveragedepth': 1,
    }

    ruleset = config_rates_ssh
    for i in ruleset.keys():
        config[i] = ruleset[i]

    # husk "ifconfig eth1 promisc", ellers blir det veldig stille.
    print "opening device %s" % config["dev"]
    p = pcapy.open_live(config["dev"], 1600, 0, 0)
    p.setfilter("udp and port 53") 
    p.setnonblock(1)

    clistore = ClientStorage(packets= config["numpackets"])
    eagle = iodClientFinderDST( config )

    try:
        p.loop(-1, packetHandler)
        print "loop returned"
    except KeyboardInterrupt:
        print "Normal exit"
    except Exception, e:
          print e
          import traceback
          traceback.print_exc(5, file=sys.stdout)
    print "dispact returned. Exiting"
