#!/usr/bin/python
# .- coding: utf-8 -.
#
# Oppsummer et datasett
# 

import os, sys
import pcapy
from impacket.ImpactDecoder import EthDecoder, LinuxSLLDecoder, UDPDecoder
from TemporalStorage import TemporalStorage, MaxSamplesReached
from iodFinder import iodFinder

class Decoder():
    def __init__(self, pcapObj):
        datalink = pcapObj.datalink()
        if pcapy.DLT_EN10MB == datalink:
            self.decoder = EthDecoder()
        elif pcapy.DLT_LINUX_SLL == datalink:
            self.decoder = LinuxSLLDecoder()
        else:
            raise Exception("Datalink type not supported: " % datalink)

        self.pcap = pcapObj
        self.connections = {}
        self.UDPDecoder = UDPDecoder()

    def packetHandler(self, hdr, data):
        global tstore, eagle, first_ts
        p = self.decoder.decode(data)

        # hopp over ipv6-pakkene, impacket st0tter det ikke.
        ptype = "%x" % ord(p.get_data_as_string()[0])
        if ptype == "60":
            return

        ip = p.child()
        try:
            # only accept udp packets
            if ip.get_ip_p() != 17:
                return 

        except Exception, e:
            print ip
            print >>sys.stderr, e
            raise Exception
            return

        udp = ip.child()
        data = udp.get_data_as_string()

        dns1 = ord(data[2]) 
        dns2 = ord(data[3])

        dnscode = "%x%x" % (dns1, dns2)

        # standard query response, no error
        if not dnscode == "8180": 
            return
        res = tstore.add( hdr, p )
        # if res returns something else than None, a new sample is ready. 
        if res:
            execute( tstore )
    def start(self):
        self.pcap.loop(0, self.packetHandler)


class iodCharacteristics(iodFinder):
    def __init__(self, slen):
        iodFinder.__init__(self, slen)

    def clistore_sum(self):
        r = {}
        for key in self.clistore.values()[0]:
            for client in self.clistore.keys():
                 try:
                     r[key] += self.clistore[client][key]
                 except KeyError:
                     r[key] = self.clistore[client][key]
        return r 

    def report(self):
        clisum = self.clistore_sum()
        sumkeys = [ "bytes","queries" ]

        print "%.0f & %.0f & %.0f B & %.0fs \\\\ \\hline " % \
            (len(self.clistore.keys()),
             clisum["queries"],
             clisum["bytes"], 
             self.sample_length, 
             )
        print

def execute( tstore ):
    global eagle
    dataset_length = tstore.range()[1] - tstore.range()[0]
    eagle.sample_length = dataset_length
    eagle.add( tstore.get_data()) 
    eagle.preprocess()
    eagle.report()
    eagle.reset()
    tstore.expunge()

if __name__ == "__main__":
    filename = sys.argv[1] 
    p = pcapy.open_offline(filename)

    import os.path
    print "%s & " % os.path.basename(filename),

    sample_length = 99999999999
    tstore = TemporalStorage(period= sample_length)
    eagle = iodCharacteristics( sample_length )

    first_ts = None

    try:
        Decoder(p).start()
        # siste rest
        execute( tstore ) 

    except MaxSamplesReached:
        pass
    except KeyboardInterrupt:
        pass
